/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

#include "extern.h"

Node *program_unit = NN;

static U_DEFINE_STACK (sblock, Node*, 20)
#define sblocktop (sblock[sblock_top-1])

/**************************** C H E C K **************************************/

Node *block_top ()
{
	return sblocktop;
}

Node *block_botton ()
{
	return sblock [0];
}

block_push (node)
Node *node;
{
	U_stack_push_sblock (node);
}

Node *block_pop ()
{
	if (sblock_top)
		return U_stack_pop_sblock ();
	else
		return NN;
}

block_add_stmt (stmt)
Node *stmt;
{
	if (stmt != &IgnoreStmt)
		block_add_sta (make_Statement (NN, stmt));
}

block_add_stmt_with_label (stmt, label, file, line)
Node *stmt, *label;
char *file;
int line;
{
	Node *sta ;

	if (stmt != &IgnoreStmt) {
		stmt = do_check_Statement (label, stmt);

		if (stmt != &IgnoreStmt) {
			sta = make_Statement (label, stmt);
			sta->Statement.file = file;
			sta->Statement.line = line;
			block_add_sta (sta);
		}
	}
}

block_add_sta (sta)
Node *sta;
{
	Node **addr_block = (Node**) sblocktop->def.trans;
	Node *block;

	if (*addr_block == NN) {
		block = make (BLOCK);
		block->Block.Statement = sta;
	}
	else {
		block = make (BLOCK);
		block->Block.Block = *addr_block;
		block->Block.Statement = sta;
	}

	*addr_block = block;

/*
	printf ("\nAdded statement: %s\n", node_Symbol (sta->Statement.Stmt->com.type));
	block_show (sta);
*/
}

/*
 * Looks for node of type 'to_this' in stack
 * If it finds it on top is OK, otherwise looks down in the stack to see if it
 * finds it. If it finds it gives as error misplaced and reorganizes the stack,
 * otherwise gives as error unmatched.
 */

Node *block_match (this, to_this)
char *this;
int to_this;
{
	Node *match = NN;

	if (sblocktop->com.type == to_this)
		match = sblocktop;
	else {
		int i, j;

		for (i = sblock_top - 2; i > 0; i--)
			if ((match = sblock[i])->com.type == to_this) {
				sprintf (serr, "misplaced %s statement", this);
				fatal (serr);

				for (j = i; j < sblock_top - 1; j++)
					sblock[j] = sblock[j + 1];

				return match;
			}

		sprintf (serr, "unmatched %s statement", this);
		fatal (serr);
	}

	return match;
}

Node *block_match_exit (name, type)
char *name;
int type;
{
	Node *match;
	int i;

	for (i = sblock_top - 1; i > 0; i--)
		if ((match = sblock[i])->com.type == DO_CONSTRUCT
		&&
		(name == CN || !strcmp (name, match->DoConstruct.DoStmt->DoStmt.Name->com.str)))
			return match;

	sprintf (serr, "unmatched %s statement", node_Symbol (type));
	fatal (serr);

	return NN;
}

Node *block_match_continue (label)
Node *label;
{
	int to_this = DO_CONSTRUCT;
	Node *match = NN;

	if (sblocktop->com.type == to_this) {
		if (sblocktop->DoConstruct.label_do != NN
		&&  !strcmp (label->com.str, sblocktop->DoConstruct.label_do->com.str))
			match = sblocktop;
	}

	return match;
}

/************************* T R A N S L A T E *********************************/

static Stack *stmt_decl, *stmt_init;

block_add_stmt_decl (stmt)
Node *stmt;
{
	if (stmt != &IgnoreStmt)
		block_add_sta_decl (make_Statement (NN, stmt));
}

block_add_sta_decl (sta)
Node *sta;
{
	stack_push (stmt_decl, sta);
}

block_push_program_unit ()
{
	program_unit = make (PROGRAM_UNIT);
	program_unit->ProgramUnit.trans = (Trans*) &program_unit->ProgramUnit.SpecificationPartBlock;
	U_stack_clear_sblock ();
	U_stack_push_sblock (program_unit);
}

block_init ()
{
	block_push_program_unit ();

	stmt_decl = stack_malloc (20);
	stmt_init = stack_malloc (20);
}

block_add_stmt_init (stmt)
Node *stmt;
{
	if (stmt != &IgnoreStmt)
		block_add_sta_init (make_Statement (NN, stmt));
}

block_add_sta_init (sta)
Node *sta;
{
	stack_push (stmt_init, sta);
}

block_add_sta_check ()
{
	int i;

	for (i = sblock_top - 1; i > 0; i--) {
		sprintf (serr, "unclosed %s", node_Symbol (sblock[i]->com.type));
		fatal (serr);
	}
}

block_add_sta_clear ()
{
	Node *top;

	if (sblock_top != 1)
		while ((top = U_stack_pop_sblock ()) != NN)
			;

	block_push_program_unit ();

	stack_clear (stmt_decl);
	stack_clear (stmt_init);
}

block_write_sta_init ()
{
	int i;

	for (i = 0; i < stmt_init->top; i++)
		write_pass ((Node*) stmt_init->stack [i]);
}

block_write_sta_decl ()
{
	int i;

	for (i = 0; i < stmt_decl->top; i++)
		write_pass ((Node*) stmt_decl->stack [i]);
}

#ifdef CHECK

block_show ()
{
	int i;
	Node *block, *sta, *temp;

	printf ("\nBLOCK_STACK: depth %d\n\n", sblock_top);

	for (i = sblock_top - 1; i >= 0; i--) {
		printf ("%2d %s:", i, node_Symbol (sblock[i]->com.type));

		block = (* ((Node**) sblock[i]->com.trans));
		for (LIST_LOOP (sta, block, temp))
			printf (" %s", node_Symbol (sta->Statement.Stmt->com.type));

		printf (".\n");
	}
		
	printf ("\n");
}

#endif
